/********************************************************HP*03-5310*********
*
* $Author: prakt_fw $
* $Date: 92/12/03 10:37:48 $
* $Revision: 1.10 $
*
****************************************************************************
*
* CONTENTS : parselib.c - library of low level parsers
*
***************************************************************************


		(C) Copyright Hewlett-Packard GmbH 1992

			  All rights reserved


     Hewlett-Packard makes no warranty of any kind with regard to this 
     software, including, but not limited to, the implied warranties of 
     merchantability and fitness for a particular purpose.
     Hewlett-Packard shall not be liable for errors contained herein or
     direct, indirect, special, incidental or consequential damages in 
     connection with the furnishing, performance or use of this material.

*/




/*-SYSTEM INCLUDES--------------------------------------------------------*/

#include <string.h>
#include <math.h>

/*-MODULE INCLUDES--------------------------------------------------------*/

#include <pc_mach.h>
#include <pc_spi.h>
#include <mecif.h>
#include <meciflib.h>
#include <database.h>
#include <pc_almgr.h>


/*-EXTERNAL FUNCTIONS-----------------------------------------------------*/

/*-EXTERNAL VARIABLES-----------------------------------------------------*/

/*-DEFINES----------------------------------------------------------------*/

#ifdef _MSC_VER
#define pow10(data) pow(10,data)
#endif

#define IS_NEGATIVE2(data) ((LOWBYTE(data) & 0x80) == 0x80)
#define NEGATIVE2_MASK 0xff000000L


/*-TYPEDEFS---------------------------------------------------------------*/

/*-GLOBAL FUNCTIONS-------------------------------------------------------*/

void lparse_cw(SpiCwFixed64 *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr);
void lparse_ws(SpiWsFixed *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr);
void lparse_nu(SpiNuFixed *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr);
void lparse_unspec(u_16 *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr);
void lparse_alstat(u_16 *rmptr,Session **session_ptr_p);
void lparse_cddnu(SpiCddNuFixed *rmptr,Session **session_ptr_p,
		 MsgIdTyp *id_ptr);
void lparse_cddset(SpiCddSetFixed *rmptr,Session **session_ptr_p,
		  MsgIdTyp *id_ptr);
void lparse_cdddev(SpiCddDevNameFixed *rmptr,Session **session_ptr_p,
		  MsgIdTyp *id_ptr);
void lparse_cddtran(SpiCddTransparent *rmptr,Session **session_ptr_p,
		   MsgIdTyp *id_ptr);

/*-GLOBAL VARIABLES-------------------------------------------------------*/

/*-STATIC FUNCTIONS-------------------------------------------------------*/

static void get_numerics(u_16 *mpb_label_ptr,u_16 *msg_end_ptr,
			 u_16 *rec_poi);

/*-STATIC VARIABLES-------------------------------------------------------*/

static Numeric		nu;
static CookedWave	cw;
static WaveSupp		ws;
static Alert		unsp;
static AlStat		alstat;
static EnumData		enum_data;
static DevName		dev_name;
static TransparentData	trans_data;


/***************************************************************************
*
* ROUTINE:lparse_cw
*
* WRITTEN BY: Juergen Barthel                         ON:05.05.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine receives the samples of a cooked wave, writes
* 	      them to a struct of type CookedWave and calls rx_lib_cw()
*
****************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
****************************************************************************/
void
lparse_cw(SpiCwFixed64 *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr)
{
cw.con_id = (*session_ptr_p)->con_id;
cw.msg_id = *id_ptr;
cw.num_of_samples = rmptr->rec_size;
memcpy(cw.samples,rmptr->sample,rmptr->rec_size * 2);
rx_lib_cw(&cw);
}

/***************************************************************************
*
* ROUTINE:lparse_ws
*
* WRITTEN BY: Juergen Barthel                         ON:05.05.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine writes received wave support data to a struct of
*		type WaveSupport and calls rx_lib_ws()
*
****************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
*
****************************************************************************/
void
lparse_ws(SpiWsFixed *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr)
{
ws.con_id     	  = (*session_ptr_p)->con_id;
ws.msg_id     	  = *id_ptr;
ws.status     	  = rmptr->psw;
ws.label_code 	  = rmptr->label_code;
ws.num_of_samples = rmptr->number_of_samples;

c16_to_c8(rmptr->label_string,ws.label,SPI_LABEL_LENGTH);

rx_lib_ws(&ws);
}

/***********************************************************************
*
* ROUTINE:lparse_nu
*
* WRITTEN BY: Juergen Barthel                   ON: 05.05.91
* CHANGED BY:
*
************************************************************************
*
* DESCRIPTION:this routine parses received numeric data and calls
*	      rx_lib_nu()
*
************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* ----------------------------------------------------------------------
*
***********************************************************************/
void
lparse_nu(SpiNuFixed *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr)
{
u_16		*mpb_label_ptr;
u_16		*rec_poi;
u_16    	*msg_end_ptr;

nu.msg_id 	 = *id_ptr;
nu.status 	 = rmptr->psw;
nu.con_id 	 = (*session_ptr_p)->con_id;
nu.label_code    = rmptr->label_code;

/* set pointer to the beginning of the nu_label in the received message */
mpb_label_ptr = &(rmptr->label_string[0]);

/* last adress of MPB NU-message */
msg_end_ptr = (u_16 *)rmptr + rmptr->mesg_length;

/* record pointer points to first optional record of NU (SPI) */
rec_poi =  (u_16 *)rmptr + 2 + rmptr->rec_size;

get_numerics(mpb_label_ptr,msg_end_ptr,rec_poi);

rx_lib_nu(&nu);
}

/***************************************************************************
*
* ROUTINE:lparse_unspec
*
* WRITTEN BY: Juergen Barthel                         ON:15.04.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine receives unspec data and calls the corresponding
*	      rx_lib_...() function
*
****************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
*
****************************************************************************/
void
lparse_unspec(u_16 *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr)
{
unsp.con_id = (*session_ptr_p)->con_id;

switch(id_ptr->ChannelId)
    {
    case CHA_AlText:
	{
	unsp.source_id  = ((AlMngrAl *)rmptr)->source_id;
	unsp.source_no  = ((AlMngrAl *)rmptr)->source_num;
	unsp.label_code = ((AlMngrAl *)rmptr)->label_code;
	unsp.status 	= ((AlMngrAl *)rmptr)->psw;
	c16_to_c8(((AlMngrAl *)rmptr)->alert_string,
				      unsp.unsptext,SPI_ALERT_LENGTH);
	rx_lib_al(&unsp);
	}
    break;

    case CHA_InText:
	{
	unsp.source_id  = ((AlMngrIn *)rmptr)->source_id;
	unsp.source_no  = ((AlMngrIn *)rmptr)->source_num;
	unsp.label_code = ((AlMngrIn *)rmptr)->label_code;
	unsp.status 	= ((AlMngrIn *)rmptr)->psw;
	c16_to_c8(((AlMngrIn *)rmptr)->alert_string,
				      unsp.unsptext,SPI_ALERT_LENGTH);
	rx_lib_inop(&unsp);
	}
    break;

    default:
    break;
    }
}

/***************************************************************************
*
* ROUTINE:lparse_alstat
*
* WRITTEN BY: Juergen Barthel                         ON:09.10.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine receives alarm status data and calls
*	      rx_lib_alstat()
*
****************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
*
****************************************************************************/
void
lparse_alstat(u_16 *rmptr,Session **session_ptr_p)
{
alstat.con_id 	     = (*session_ptr_p)->con_id;
alstat.al_severity   = (u_8)(((AlMngrAs *)rmptr)->al_status1 & AL_STAT1_ALARM_MASK);
alstat.inop_severity = (u_8)(((AlMngrAs *)rmptr)->al_status1 & AL_STAT1_INOP_MASK);
alstat.mon_status    =
	(u_8)((((AlMngrAs *)rmptr)->unit_cap & UNIT_CAP_MONIT_STATUS) >> 11);

if( (((AlMngrAs *)rmptr)->al_status3) & AL_STAT3_MAIN_AL_OFF )
    {
    /* main alarms are suspended */
    alstat.al_suspended = 1;
    }
else
    {
    /* main alarms are on */
    alstat.al_suspended = 0;
    }

rx_lib_alstat(&alstat);
}

/***********************************************************************
*
* ROUTINE:lparse_cddnu
*
* WRITTEN BY: Juergen Barthel                   ON: 15.11.91
* CHANGED BY:
*
************************************************************************
*
* DESCRIPTION:this parses received cdd numeric data and calls rx_lib_nu()
*
************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* ----------------------------------------------------------------------
*
***********************************************************************/
void
lparse_cddnu(SpiCddNuFixed *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr)
{
u_16		*mpb_label_ptr;
u_16		*rec_poi;
u_16    	*msg_end_ptr;

nu.msg_id 	 = *id_ptr;
nu.status 	 = rmptr->validity;
nu.con_id 	 = (*session_ptr_p)->con_id;
nu.label_code    = rmptr->label_code;

/* set pointer to the beginning of the nu_label in the received message */
mpb_label_ptr = &(rmptr->label_string[0]);

/* last adress of MPB NU-message */
msg_end_ptr = (u_16 *)rmptr + rmptr->mesg_length;

/* record pointer points to first optional record of NU (SPI) */
rec_poi =  (u_16 *)rmptr + 2 + rmptr->rec_size;

get_numerics(mpb_label_ptr,msg_end_ptr,rec_poi);

rx_lib_nu(&nu);
}

/***********************************************************************
*
* ROUTINE:lparse_cddset
*
* WRITTEN BY: Juergen Barthel                   ON: 19.11.91
* CHANGED BY:
*
************************************************************************
*
* DESCRIPTION:this routine parses received cdd setting data and calls
*	      the corresponding rx_lib_...() function
*
************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* ----------------------------------------------------------------------
*
***********************************************************************/
void
lparse_cddset(SpiCddSetFixed *rmptr,Session **session_ptr_p,MsgIdTyp *id_ptr)
{
SpiCddEnum	*enum_ptr;
u_16		*mpb_label_ptr;
u_16		*rec_poi;
u_16    	*msg_end_ptr;


/* search for the first optional record with ID SPI_NUMERIC or SPI_ENUM */

/* record pointer points to first optional record of NU (SPI) */
rec_poi =  (u_16 *)rmptr + 2 + rmptr->rec_size;

/* last adress of MPB NU-message */
msg_end_ptr = (u_16 *)rmptr + rmptr->mesg_length;

while (rec_poi <= msg_end_ptr)
    {
    if (LOWBYTE(*rec_poi)== SPI_NUMERIC)
	{
	/* appended record contains numerics --> store them *
	 * and call rx_lib_nu()				    */
	mpb_label_ptr    = rmptr->label_string;
	nu.con_id 	 = (*session_ptr_p)->con_id;
	nu.msg_id 	 = *id_ptr;
	nu.status 	 = 0;
	nu.label_code    = rmptr->label_code;
	get_numerics(mpb_label_ptr,msg_end_ptr,rec_poi);
	rx_lib_nu(&nu);
	return;
	}
    else  if (LOWBYTE(*rec_poi)== SPI_ENUM)
	{
	/* appended record contains enum data --> store it *
	 * and call rx_lib_enum()			   */
	enum_ptr = (SpiCddEnum *)rec_poi;
	enum_data.con_id    = (*session_ptr_p)->con_id;
	enum_data.msg_id    = *id_ptr;
	enum_data.enum_code = enum_ptr->enum_code;
	enum_data.label_code = rmptr->label_code;
	c16_to_c8(rmptr->label_string, enum_data.label, SPI_LABEL_LENGTH);

	c16_to_c8(enum_ptr->enum_string,
		  enum_data.enum_string,SPI_ENUM_LENGTH);
	rx_lib_enum(&enum_data);
	return;
	}
    /* next optional record */
    rec_poi = rec_poi + HIGHBYTE(*rec_poi) + 1;
    }
}

/***********************************************************************
*
* ROUTINE:lparse_cdddev
*
* WRITTEN BY: Juergen Barthel                   ON: 19.11.91
* CHANGED BY:
*
************************************************************************
*
* DESCRIPTION:this routine parses received cdd device name data and calls
*	      rx_lib_devname()
*
************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* ----------------------------------------------------------------------
*
***********************************************************************/
void
lparse_cdddev(SpiCddDevNameFixed *rmptr,Session **session_ptr_p,
	     MsgIdTyp *id_ptr)
{
dev_name.con_id = (*session_ptr_p)->con_id;
dev_name.msg_id = *id_ptr;

/* copy the device name to the dev_name struct */
c16_to_c8(rmptr->dev_name,dev_name.dev_name,CDD_DEV_NAME_LENGTH);

rx_lib_devname(&dev_name);
}

/***********************************************************************
*
* ROUTINE:lparse_cddtran
*
* WRITTEN BY: Juergen Barthel                   ON: 19.11.91
* CHANGED BY:
*
************************************************************************
*
* DESCRIPTION:this routine parses received cdd transparent data and calls
*	      rx_lib_trans()
*
************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* ----------------------------------------------------------------------
*
***********************************************************************/
void
lparse_cddtran(SpiCddTransparent *rmptr,Session **session_ptr_p,
	       MsgIdTyp *id_ptr)
{
trans_data.con_id 	= (*session_ptr_p)->con_id;
trans_data.msg_id 	= *id_ptr;
trans_data.num_of_words = rmptr->rec_size;

/* copy the transparent data */
memcpy(rmptr->data_word,trans_data.data_word,trans_data.num_of_words);

rx_lib_transp(&trans_data);
}

/***********************************************************************
*
* ROUTINE:get_numerics
*
* WRITTEN BY: Juergen Barthel                   ON: 15.11.91
* CHANGED BY:
*
************************************************************************
*
* DESCRIPTION:this routine writes received numeric data to a struct
*		of type Numeric, it is assumed that nu.status was set
*		previously
*
************************************************************************
* RETURNED VALUE:pointer to a struct of type Numeric
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* ----------------------------------------------------------------------
*
***********************************************************************/
static void
get_numerics(u_16 *mpb_label_ptr,u_16 *msg_end_ptr,u_16 *rec_poi)
{
SpiNumeric	*numeric_poi;
i_16		exponent;
i_32		mantissa;

c16_to_c8(mpb_label_ptr,nu.label,SPI_LABEL_LENGTH);

/* parse all SpiNumeric records until end of message */
nu.num_of_values  = 0;
while (rec_poi <= msg_end_ptr)
    {
    if (LOWBYTE(*rec_poi)== SPI_NUMERIC)
	{
	/* overlay NU numeric structure */
	numeric_poi = (SpiNumeric *)rec_poi;

	/* get the numeric value,remember: the value is in DNF-format (SPI)*/
	mantissa = LOWBYTE(numeric_poi->value_1);
	/* 24 bit numeric */
	mantissa = (mantissa << 16) + numeric_poi->value_2;

	/* make the right sign for the mantissa */
	if (IS_NEGATIVE2(numeric_poi->value_1))
	    {
	    mantissa |= NEGATIVE2_MASK;
	    }

	/* make the right sign for the exponent */
	exponent = HIGHBYTE(numeric_poi->value_1);
	if (exponent & 0x080)
	    {
	    exponent |= 0xff00;
	    }

	if (((nu.status & 0X0080) == 0X0080) || (exponent > 10))
	    {
	    nu.numeric[nu.num_of_values].value = 0;
	    }
	else
	    {
	    nu.numeric[nu.num_of_values].value =
					((float)mantissa) * pow10(exponent);
	    }
	nu.numeric[nu.num_of_values].label_code = numeric_poi->label_code;
	(nu.num_of_values)++;
	}

    /* next optional record */
    rec_poi = rec_poi + HIGHBYTE(*rec_poi) + 1;
    }
}
