/********************************************************HP*03-5310*********
*
* $Author: prakt_fw $
* $Date: 92/12/03 10:37:53 $
* $Revision: 1.10 $
*
*
*
****************************************************************************
*
* CONTENTS : rs232.c - This module conducts the RS232 in the VECTRA PC
*
***************************************************************************


		(C) Copyright Hewlett-Packard GmbH 1992

			  All rights reserved


     Hewlett-Packard makes no warranty of any kind with regard to this
     software, including, but not limited to, the implied warranties of
     merchantability and fitness for a particular purpose.
     Hewlett-Packard shall not be liable for errors contained herein or
     direct, indirect, special, incidental or consequential damages in
     connection with the furnishing, performance or use of this material.

*/



/*-SYSTEM INCLUDES--------------------------------------------------------*/

#include <stdlib.h>


#ifdef _MSC_VER   /* Microsoft C directive            */

   #include <conio.h>

   void _disable(void);
   void _enable(void);

   #define  asm   _asm
#else            /* TURBO C directive */
   #pragma inline
#endif



#ifdef _MSC_VER 
   #ifndef SLOW
   #define  MSC_VOID_PTR
   void (  *_dos_getvect (unsigned))();
   void _dos_setvect (unsigned, void (  *)());
   #else
   #include <dos.h>
   #endif
#else
   #include <dos.h>
   #undef   MSC_VOID_PTR
#endif





/*-MODULE INCLUDES--------------------------------------------------------*/

#include <pc_mach.h>
#include <meciflib.h>

/*-EXTERNAL FUNCTIONS-----------------------------------------------------*/

extern	void	rx_error(u_16 error_code, i_16 con_id);


/*-EXTERNAL VARIABLES-----------------------------------------------------*/

extern PortTbl	confport_tbl[];
extern Session  *session_ptr[];


/*-DEFINES----------------------------------------------------------------*/


#ifdef SLOW
   #define C_PORT_ISR(N)                                \
register	iir;                                 \
						     \
Repeat:				                     \
iir = inportb(IIR##N) & 0x07;                        \
if (iir == 0x04)                                     \
    {                                                \
    INT_FIFO_WRITE((&inputfifo[N]),inportb(RBR##N ));\
    if (inputfifo[N].limit == inputfifo[N].content)  \
	outportb(MCR##N ,inportb(MCR##N ) & DTR_off);\
    goto Repeat;                                     \
    }                                                \
else if (iir == 0x02)                                               \
    {                                                               \
    /* if there is anything in the outputbuffer,try to transmit it*/\
    if (outputfifo[N].content)                                      \
	{                                                           \
	/* check CTS = ON and DSR = ON                              \
	 * if TRUE: transmit next byte in outputbuffer *            \
	 * else ERROR          */                                   \
	if ((inportb(MSR##N ) & CTS_DSR ) == CTS_DSR)               \
	    {                                                       \
	    outportb(THR##N ,*(outputfifo[N].read_ptr));            \
	    (outputfifo[N].read_ptr)++;                             \
	    (outputfifo[N].content)--;                              \
								    \
	    /* check read_ptr pointer position */                   \
	    if(outputfifo[N].read_ptr > outputfifo[N].end)          \
		outputfifo[N].read_ptr = outputfifo[N].start;       \
	    hw_hs_wait[N] = 0;                                      \
	    }                                                       \
	else                                                        \
	    {                                                       \
	    hw_hs_wait[N] = 1;                                      \
	    }                                                       \
	}                                                           \
    else                                                            \
	outportb(MCR##N ,inportb(MCR##N ) & RTS_off);               \
    goto Repeat;                                                    \
    }                                                               \



#define C_END_MASTER_ISR(N)                               \
/* send END OF INTERRUPT to master interrupt controller */\
outportb(MEOI,EOI##N );


#define C_END_SLAVE_ISR(N) \
/* send END OF INTERRUPT to master interrupt controller */\
outportb(MEOI,EOSI);                                      \
							  \
/* send END OF INTERRUPT to slave interrupt controller */ \
outportb(SEOI,EOI##N );


#define C_MASTER_PORT_ISR(N) \
C_PORT_ISR(N)		     \
C_END_MASTER_ISR(N)


#define C_SLAVE_PORT_ISR(N) \
C_PORT_ISR(N)               \
C_END_SLAVE_ISR(N)

#endif



#define FIFO_INIT(name,buf_name,buf_size)		\
		{                              		\
		name->start      = buf_name;        	\
		name->end        = buf_name + buf_size -1;\
		name->write_ptr = buf_name;       	\
		name->read_ptr  = buf_name;        	\
		name->content    = 0;                 	\
		name->size       = buf_size;            \
		name->limit	 = buf_size - 3;	\
		}

#define FIFO_FLASH(name)	                	\
		{                              		\
		name->write_ptr = name->start;       	\
		name->read_ptr  = name->start;        	\
		name->content    = 0;                 	\
		}


#define FIFO_READ(name,c)				\
		{          				\
		c = *(name->read_ptr); 		        \
		(name->read_ptr)++;    		        \
		disable();              		\
		(name->content)--;      		\
		enable();				\
		/* check read_ptr pointer position */ 	\
		if(name->read_ptr > name->end)   	\
		    name->read_ptr = name->start;	\
		}


#define FIFO_WRITE(name,element)			\
		{                                	\
		*(name->write_ptr) = element;   	\
		(name->write_ptr)++;            	\
		disable();                       	\
		(name->content)++;               	\
		enable();                        	\
		/* check writepointer position      */	\
		if(name->write_ptr > name->end)      	\
		    name->write_ptr = name->start;   	\
		}

#define INT_FIFO_WRITE(name,element)			\
		{                                	\
		*(name->write_ptr) = element;   	\
		(name->write_ptr)++;            	\
		(name->content)++;               	\
		/* check writepointer position      */	\
		if(name->write_ptr > name->end)      	\
		    name->write_ptr = name->start;   	\
		}

#define PORT_CONFIG_ADD 0X400 /* BIOS address including port configuration */
#define DLL sio[port]               /* addresses of the UART registers           */
#define DLM (sio[port] + 1)
#define RBR sio[port]
#define THR sio[port]
#define IER (sio[port] + 1)
#define IIR (sio[port] + 2)
#define LCR (sio[port] + 3)
#define MCR (sio[port] + 4)
#define LSR (sio[port] + 5)
#define MSR (sio[port] + 6)
#define SCR (sio[port] + 7)
#define RBR0 0X3F8
#define THR0 0X3F8
#define IIR0 0X3FA
#define MCR0 0X3FC
#define LSR0 0X3FD
#define MSR0 0X3FE
#define RBR1 0X2F8
#define THR1 0X2F8
#define IIR1 0X2FA
#define MCR1 0X2FC
#define LSR1 0X2FD
#define MSR1 0X2FE
#define RBR2 0X3E8
#define THR2 0X3E8
#define IIR2 0X3EA
#define MCR2 0X3EC
#define LSR2 0X3ED
#define MSR2 0X3EE
#define RBR3 0X2E8
#define THR3 0X2E8
#define IIR3 0X2EA
#define MCR3 0X2EC
#define LSR3 0X2ED
#define MSR3 0X2EE

#define MEOI 0X20             /* address of master interrupt controller */
#define SEOI 0XA0             /* address of slave interrupt controller  */

#define EOI0 0X64              /* EOI command for port 0 */
#define EOI1 0X63              /* EOI command for port 1 */
#define EOSI 0X62              /* EOI command for slave interrupt */

#define EOI2 0X62              /* EOI command for port 2 */
#define EOI3 0X63              /* EOI command for port 3 */

#define INBUFSIZE   4096
#define OUTBUFSIZE  4096
#define MCR_int_en  0X08
#define DTR_on      0X01 		/* mask to set DTR = ON  */
#define DTR_off     0XFE		/* mask to set DTR = OFF */
#define RTS_on      0X02		/* mask to set RTS = ON  */
#define RTS_off     0XFD		/* mask to set RTS = OFF */
#define CTS_DSR     0X30
#define	OVERFL_BIT   0X02



/*-TYPEDEFS---------------------------------------------------------------*/

/* typedefinition of a software fifo to store incomming and outgoing data */
/* ATTENTION: if this structure is to be changed remember the effects on  *
 * 	      the inerrupt service routines and the inline assembler code!*/
typedef struct {
u_8 	*start;
u_8 	*end;
u_8 	*write_ptr;
u_8 	*read_ptr;
i_16    content;
i_16    size;
i_16	limit;
} Fifo8;


/*-GLOBAL FUNCTIONS-------------------------------------------------------*/

i_16   	mecif_init(u_16 port,u_16 baudrate);
i_16   	write_msg(u_16 port,u_16 *mptr);
u_16  	*read_msg(u_16 port);
void  	mecif_end(u_16 port);


/*-GLOBAL VARIABLES-------------------------------------------------------*/

/* if type u_16 is changed remember effect on the inline assembly code */
u_16 	lost_char[MAX_HW_PORTS];
u_16	over_flag[MAX_HW_PORTS];

/*-STATIC FUNCTIONS-------------------------------------------------------*/

#ifdef SLOW
static void interrupt	mecif0_isr(void);         /* MECIF ISR of port 0 */
static void interrupt 	mecif1_isr(void);         /* MECIF ISR of port 1 */
static void interrupt 	mecif2_isr(void);         /* MECIF ISR of port 2 */
static void interrupt 	mecif3_isr(void);         /* MECIF ISR of port 3 */
#else
static void mecif0_isr(void);         /* MECIF ASSEMBLER ISR of port 0 */
static void mecif1_isr(void);         /* MECIF ASSEMBLER ISR of port 1 */
static void mecif2_isr(void);         /* MECIF ASSEMBLER ISR of port 2 */
static void mecif3_isr(void);         /* MECIF ASSEMBLER ISR of port 3 */
#endif


#ifndef MSC_VOID_PTR

static void near    	new_isr(u_16 port,void (interrupt *isr)(void));
static void (interrupt  *old_isr[MAX_HW_PORTS])();

#else

static void near    	new_isr(u_16 port,void ( *isr)(void));
static void (  *old_isr[MAX_HW_PORTS])();

#endif


static void near    	restore_old_isr(u_16 port);

/*-STATIC VARIABLES-------------------------------------------------------*/

static u_8 near 	old_IER[MAX_HW_PORTS];
static u_8 near 	old_MCR[MAX_HW_PORTS];
static u_8 near 	old_LCR[MAX_HW_PORTS];
static u_8 near 	old_msb[MAX_HW_PORTS];
static u_8 near 	old_lsb[MAX_HW_PORTS];
static u_16 near 	sio[MAX_HW_PORTS];        /* address of UART port */


static u_8 near         hw_hs_wait[MAX_HW_PORTS];
static u_16 near  	length[MAX_HW_PORTS];
static u_8 near  	sync_flag[MAX_HW_PORTS];
static u_8 near 	rm[MAX_HW_PORTS][MAX_MSG_LENGTH+2];
static Fifo8 near		inputfifo[MAX_HW_PORTS];
static Fifo8 near		outputfifo[MAX_HW_PORTS];
static u_8    		*rsm[MAX_HW_PORTS];

/*  seriell IO interrupt vector numbers  */
static const u_8 near	int_vect[] = {0X0C, 0X0B, 0X72, 0X73};

/*  interrupt mask register addresses */
static const i_16 near	imr_addr[] = {0X21, 0X21, 0XA1, 0XA1};

/*  masks for enabling SIO interrupts  */
static const u_8 near	unmask_int[] = {0XEF, 0XF7, 0XFB, 0XF7};

/*  masks for disabling SIO interrupts */
static const u_8 near   mask_int[] = {0X10, 0X08, 0X04, 0X08};


/***************************************************************************
*
* ROUTINE:mecif_init
*
* WRITTEN BY: Juergen Barthel                         ON:13.3.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:initializes the MECIF interface
*
****************************************************************************
* RETURNED VALUE: SUCCESS on success, else
*
*		  FAIL if one of the parameters passed to mecif_init is
*		       invalid,
*
*		  NOT_ENOUGH_MEMORY if out of memory.
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* inputfifo[]      r   inputfifo structure
* outputfifo[]     r   outputfifo structure
* sio[]           r/w  port address of the SIO
* PORT_CONFIG_ADD  r   address where BIOS stores the port hardware configuration
* int_vect[]       r   SIO interrupt vector numbers
* imr_addr[]      r/w  interrupt mask register address
* mask_int[]       r   mask to disable SIO interrupts
* unmask_int[]     r   mask to enable SIO interrupts
* old_IER[]        w   stores the old content of the IER
* old_MCR[]        w   stores the old content of the MCR
* old_LCR[]        w   stores the old content of the LCR
* old_msb[]        w   stores the old content of the DLM
* old_lsb[]        w   stores the old content of the DLL
* MCR_int_en       r   mask to enable the interrupts specified in the IER
****************************************************************************/
i_16 mecif_init(u_16 port,u_16 baudrate)
{
#ifndef MSC_VOID_PTR
   static void (interrupt *dummy_isr)();
#else
   static void ( *dummy_isr)();
#endif


u_8		*inbuf_ptr;
u_8		*outbuf_ptr;
i_16     	i;
u_8		msb = 0X00;
u_8		lsb = 0X00;

rsm[port]    = &(rm[port][0]);
length[port] = 0;

if ((inbuf_ptr = (u_8 *)malloc(INBUFSIZE)) == NULL)
    {
    return(NOT_ENOUGH_MEMORY);
    }

if ((outbuf_ptr = (u_8 *)malloc(OUTBUFSIZE)) == NULL)
    {
    free(inbuf_ptr);
    return(NOT_ENOUGH_MEMORY);
    }

FIFO_INIT((&inputfifo[port]),inbuf_ptr,INBUFSIZE);
FIFO_INIT((&outputfifo[port]),outbuf_ptr,OUTBUFSIZE);

lost_char[port] = 0;
over_flag[port] = 0;
sync_flag[port] = FALSE;

sio[port] = *((u_16 far *)PORT_CONFIG_ADD + port );

/*  is port not available ?  */
if ((port >= MAX_HW_PORTS) || (sio[port] == 0))
    {
    free(inbuf_ptr);
    free(outbuf_ptr);
    return(PORT_WRONG);
    }
else
    {
    /*  mask bitN in the interrupt controller *
     *  and disable interrupts in the RS232   */
    outportb(imr_addr[port],inportb(imr_addr[port]) | mask_int[port]);
    old_MCR[port] = inportb(MCR);
    old_LCR[port] = inportb(LCR);
    old_IER[port] = inportb(IER);          /* saving the old IER */

    /*  initialize COMport baudrate : 9600, 19200 or 38400 */
    if (baudrate == 9600)
	lsb = 0X0C;
    else if (baudrate == 19200)
	lsb = 0X06;
    else if (baudrate == 38400U)
	lsb = 0X03;

    if (lsb == 0X00)
	{
	free(inbuf_ptr);
	free(outbuf_ptr);
	return(BAUDRATE_WRONG);
	}
    else
	{
	outportb(LCR,0X80);
	old_lsb[port] = inportb(DLL);
	old_msb[port] = inportb(DLM);
	outportb(DLL,lsb);
	outportb(DLM,msb);

       /* initialize COMport	# of data bits	:   8 	*
	* initialize COMport	# of stop bits	:   1   *
	* initialize COMport	parity		:  none	*/
	outportb(LCR,0X03);

       /* set interrupt vector to MECIF isr */
	switch(port)
	    {
	    case 0:
		dummy_isr = mecif0_isr;
	    break;
	    case 1:
		dummy_isr = mecif1_isr;
	    break;

	    case 2:
		dummy_isr = mecif2_isr;
	    break;

	    case 3:
		dummy_isr = mecif3_isr;
	    break;
	    }
	new_isr(port,dummy_isr);

       /* set interrupt when THR runs empty(ETBEI)   *
	* and interrupt when data is received (ELSI) */
	outportb(IER,0X03);

       /* setting the MCR to enable interrupts in IER */
	outportb(MCR,MCR_int_en);

       /* dummy read of SIO registers */
	for (i = 0;i <= 6;i++)
	    inportb(sio[port] + i);
	outportb(imr_addr[port],inportb(imr_addr[port]) & unmask_int[port]);

	outportb(MCR,((inportb(MCR) | DTR_on) & RTS_off));/* DTR=ON,RTS=OFF */
	return(SUCCESS);
	}
    }
}

/***************************************************************************
*
* ROUTINE:new_isr
*
* WRITTEN BY: Juergen Barthel                         ON:5.3.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION: this routine sets the interrupt vector # int_vect[port] to
* 		mecif_isr, it is necessary because the function 'setvect'
*		seems to cause register troubles
*
****************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* int_vect[]       r   interrupt# of selected port
* old_isr[]        w   stores the old isr vector
****************************************************************************/
#ifndef MSC_VOID_PTR
static void near  new_isr(u_16 port,void (interrupt *isr)(void))
#else
static void near  new_isr(u_16 port,void ( *isr)(void))
#endif

{
old_isr[port] = getvect(int_vect[port]);           /* save old isr vector */
setvect(int_vect[port],isr);
}

/***************************************************************************
*
* ROUTINE:restore_old_isr
*
* WRITTEN BY: Juergen Barthel                         ON:5.3.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION: this routine installs the old ISR vector again
*
****************************************************************************
* RETURNED VALUE:none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* int_vect[]       r   interrupt # of selected port
* old_isr[]        r   contains the old isr vector
****************************************************************************/
static void near  restore_old_isr(u_16 port)
{
setvect(int_vect[port],old_isr[port]);
}

/***************************************************************************
*
* ROUTINE: mecif0_isr
*
* WRITTEN BY: Juergen Barthel                         ON:24.02.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION: this routine is the MECIF interrupt service routine for port 1
*
****************************************************************************
* RETURNED VALUE: none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* inputfifo[]      r   inputfifo structure
* outputfifo[]     r   outputfifo structure
*
****************************************************************************/
#ifdef SLOW
static void interrupt	mecif0_isr(void)
{
C_MASTER_PORT_ISR(0)
}
#else
#define start +0
#define end +4
#define write_ptr +8
#define read_ptr +12
#define content +16
#define size +18
#define limit +20


static void  mecif0_isr(void)
{
asm	push	ax			/*save all used registers,*/
asm	push 	bx			/*remember Flags are saved*/
asm	push 	dx			/*automatically by the CPU*/

asm	push 	ds
asm	push 	es



/* load correct datasegment -> DS --------------------------------------*/
   asm	mov 	dx,seg inputfifo
   asm	mov 	ds,dx


/* identificate interrupt ----------------------------------------------*/

NEXT_IIR:
asm	mov 	dx,IIR0			/*load IIR port address*/
asm	in  	al,dx			/*read IIR*/
asm	and 	al,7
asm	cmp	al,4			/*RX interrupt ?*/
asm	je	RX_INT
asm	cmp 	al,2			/*TX interrupt ?*/
asm	je	TX_INT
asm	clc	/*dummy instruction to speed up jmp instruction*/
asm	jmp	END_INT			/*prepare to leave interrupt*/
asm	clc	/*dummy instruction to speed up jmp instruction*/


/* RX interrupt handling -----------------------------------------------*/

RX_INT:
asm	mov	dx,RBR0			/*load RBR port adress*/
asm	in	al,dx			/*read received byte from UART*/

asm	les	bx,dword ptr inputfifo write_ptr /*load write pointer*/
asm	mov	byte ptr es:[bx],al	/*store received byte in fifo*/

asm	inc	word ptr inputfifo write_ptr	/*increment write pointer*/
asm	inc	word ptr inputfifo content	/*increment fifo content*/

asm	cmp	bx,word ptr inputfifo end /*wrap around ?*/
asm	je	rxm2			/*if yes handle it at rxm2*/

rxm1:
asm	mov	ax,inputfifo limit	/*load fill limit*/
asm	cmp	inputfifo content,ax /*compare fifo content vs limit*/
asm	jb	NEXT_IIR		/*if still enougth space go on*/

/* fifo full handling*/
asm	mov	dx,MCR0			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,DTR_off		/*mask HW handshake bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp short NEXT_IIR		/*see if there is another IRQ*/

/* wrap around handling*/
rxm2:
asm	mov	ax,word ptr inputfifo start /*fifo start addr. offs.->ax*/
asm	mov	word ptr inputfifo write_ptr,ax /*wrap around*/
asm	jmp short rxm1				/*jump back to continue*/


/* TX interrupt handling -----------------------------------------------*/

TX_INT:                             /*check outputfifo content*/
#ifdef _MSC_VER                     
asm   mov ax, 0                     /* ensure a word alignment for MSC  */
asm	cmp	outputfifo content, ax
#else
asm	cmp	outputfifo content, 0	
#endif

asm	je	txm2			/*if there is more to transmit*/
					/*transmit it, else set RTS off*/
asm	mov	dx,MSR0			/*load MSR port address*/
asm	in	al,dx			/*read MSR*/

asm	and	al,CTS_DSR		/*check HW handshake*/
asm	jne	txm3

asm	les	bx,dword ptr outputfifo read_ptr /*load address of next*/
							/*byte*/

asm	mov	al,byte ptr es:[bx]	/*read next byte from fifo*/

asm	mov	dx,THR0			/*load THR port address*/
asm	out	dx,al			/*write next byte to THR*/

asm	inc	word ptr outputfifo read_ptr	/*increment read pointer*/
asm	dec	word ptr outputfifo content	/*decrement fifo content*/

asm	cmp	bx,word ptr outputfifo end /*wrap around ?*/
asm	je	txm4			/*if yes handle it at txm4*/

txm1:
asm	mov	hw_hs_wait,0		/*set HW handshake wait flag*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/

/* end of transmission handling*/
txm2:
asm	mov	dx,MCR0			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,RTS_off		/*mask RTS bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/

/* missing HW handshake handling*/
txm3:
asm	mov	hw_hs_wait,1		/*set HW handshake wait flag*/
asm	jmp 	NEXT_IIR

/* wrap around handling*/
txm4:
asm	mov	ax,word ptr outputfifo start /*fifo start addr.offs.->ax*/
asm	mov	word ptr outputfifo read_ptr,ax /*wrap around*/
asm	jmp short txm1				/*jump back to continue*/

/* exit interrupt handling ---------------------------------------------*/

END_INT:
/*asm	mov	dx,MEOI			master int. cont. port addr.*/
asm	mov	al,EOI0			/*load end of interrupt command*/
asm	out	MEOI,al			/*send end of int.-> master cont.*/


asm	pop 	es			/*restore all used registers,*/
asm	pop 	ds			/*remember Flags are restored*/
asm	pop 	dx			/*automatically by the CPU*/
asm	pop 	bx
asm	pop 	ax

#ifdef _MSC_VER
   asm   pop   bp    /* restore the stack frame (changed BS)   */
#endif

asm	iret
}

#undef start
#undef end
#undef write_ptr
#undef read_ptr
#undef content
#undef size
#undef limit
#endif

/***************************************************************************
*
* ROUTINE: mecif1_isr
*
* WRITTEN BY: Juergen Barthel                         ON:24.02.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine is the MECIF interrupt service routine for port 2
*
****************************************************************************
* RETURNED VALUE: none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* inputfifo[]      r   pointer to inputfifo structure
* outputfifo[]     r   pointer to outputfifo structure
*
****************************************************************************/
#ifdef SLOW
static void interrupt	mecif1_isr(void)
{
C_MASTER_PORT_ISR(1)
}
#else
#define start +22
#define end +26
#define write_ptr +30
#define read_ptr +32
#define content +38
#define size +40
#define limit +42

static void mecif1_isr(void)
{
asm	push	ax			/*save all used registers,*/
asm	push 	bx			/*remember Flags are saved*/
asm	push 	dx			/*automatically by the CPU*/
asm	push 	ds
asm	push 	es



/* load correct datasegment -> DS --------------------------------------*/

asm	mov 	dx,seg inputfifo
asm	mov 	ds,dx


/* identificate interrupt ----------------------------------------------*/

NEXT_IIR:
asm	mov 	dx,IIR1			/*load IIR port address*/
asm	in  	al,dx			/*read IIR*/
asm	and 	al,7
asm	cmp	al,4			/*RX interrupt ?*/
asm	je	RX_INT
asm	cmp 	al,2			/*TX interrupt ?*/
asm	je	TX_INT
asm	clc	/*dummy instruction to speed up jmp instruction*/
asm	jmp	END_INT			/*prepare to leave interrupt*/
asm	clc	/*dummy instruction to speed up jmp instruction*/


/* RX interrupt handling -----------------------------------------------*/

RX_INT:
asm	mov	dx,RBR1			/*load RBR port adress*/
asm	in	al,dx			/*read received byte from UART*/

asm	les	bx,dword ptr inputfifo write_ptr/*load write pointer*/
asm	mov	byte ptr es:[bx],al	/*store received byte in fifo*/

asm	inc	word ptr inputfifo write_ptr/*increment write pointer*/
asm	inc	word ptr inputfifo content/*increment fifo content*/

asm	cmp	bx,word ptr inputfifo end/*wrap around ?*/
asm	je	rxm2			/*if yes handle it at rxm2*/

rxm1:
asm	mov	ax,inputfifo limit/*load fill limit*/
asm	cmp	inputfifo content,ax /*compare fifo content vs limit*/
asm	jb	NEXT_IIR		/*if still enougth space go on*/

/* fifo full handling*/
asm	mov	dx,MCR1			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,DTR_off		/*mask HW handshake bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp short NEXT_IIR		/*see if there is another IRQ*/

/* wrap around handling*/
rxm2:
asm	mov	ax,word ptr inputfifo start/*fifo start addr. offs.->ax*/
asm	mov	word ptr inputfifo write_ptr,ax /*wrap around*/
asm	jmp short rxm1				/*jump back to continue*/



/* TX interrupt handling -----------------------------------------------*/

TX_INT:                             /*check outputfifo content*/
#ifdef _MSC_VER                     
asm   mov ax, 0                     /* ensure a word alignment for MSC  */
asm	cmp	outputfifo content, ax
#else
asm	cmp	outputfifo content, 0
#endif

asm	je	txm2			/*if there is more to transmit*/
					/*transmit it, else set RTS off*/
asm	mov	dx,MSR1			/*load MSR port address*/
asm	in	al,dx			/*read MSR*/

asm	and	al,CTS_DSR		/*check HW handshake*/
asm	jne	txm3

asm	les	bx,dword ptr outputfifo read_ptr /*load address of next*/
							/*byte*/

asm	mov	al,byte ptr es:[bx]	/*read next byte from fifo*/

asm	mov	dx,THR1			/*load THR port address*/
asm	out	dx,al			/*write next byte to THR*/

asm	inc	word ptr outputfifo read_ptr	/*increment read pointer*/
asm	dec	word ptr outputfifo content	/*decrement fifo content*/

asm	cmp	bx,word ptr outputfifo end /*wrap around ?*/
asm	je	txm4			/*if yes handle it at txm4*/

txm1:
asm	mov	hw_hs_wait+1,0		/*set HW handshake wait flag*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/


/* end of transmission handling*/
txm2:
asm	mov	dx,MCR1			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,RTS_off		/*mask RTS bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/

/* missing HW handshake handling*/
txm3:
asm	mov	hw_hs_wait+1,1		/*set HW handshake wait flag*/
asm	jmp 	NEXT_IIR

/* wrap around handling*/
txm4:
asm	mov	ax,word ptr outputfifo start /*fifo start addr.offs.->ax*/
asm	mov	word ptr outputfifo read_ptr,ax /*wrap around*/
asm	jmp short txm1				/*jump back to continue*/


/* exit interrupt handling ---------------------------------------------*/

END_INT:
asm	mov	al,EOI1			/*load end of interrupt command*/
asm	out	MEOI,al			/*send end of int.-> master cont.*/

asm	pop 	es			/*restore all used registers,*/
asm	pop 	ds			/*remember Flags are restored*/
asm	pop 	dx			/*automatically by the CPU*/
asm	pop 	bx
asm	pop 	ax

#ifdef _MSC_VER      
   asm   pop   bp    /* restore the stack frame (changed BS)   */
#endif

asm	iret
}
#undef start
#undef end
#undef write_ptr
#undef read_ptr
#undef content
#undef size
#undef limit


#endif

/***************************************************************************
*
* ROUTINE: mecif2_isr
*
* WRITTEN BY: Juergen Barthel                         ON:24.02.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine is the MECIF interrupt service routine for port 3
*
****************************************************************************
* RETURNED VALUE: none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* inputfifo[]      r   pointer to inputfifo structure
* outputfifo[]     r   pointer to outputfifo structure
*
****************************************************************************/
#ifdef SLOW
static void interrupt	mecif2_isr(void)
{
C_SLAVE_PORT_ISR(2)
}
#else
#define start +44
#define end +48
#define write_ptr +52
#define read_ptr +56
#define content +60
#define size +62
#define limit +64

static void mecif2_isr(void)
{
asm	push	ax			/*save all used registers,*/
asm	push 	bx			/*remember Flags are saved*/
asm	push 	dx			/*automatically by the CPU*/
asm	push 	ds
asm	push 	es



/* load correct datasegment -> DS --------------------------------------*/

asm	mov 	dx,seg inputfifo
asm	mov 	ds,dx



/* identificate interrupt ----------------------------------------------*/

NEXT_IIR:
asm	mov 	dx,IIR2			/*load IIR port address*/
asm	in  	al,dx			/*read IIR*/
asm	and 	al,7
asm	cmp	al,4			/*RX interrupt ?*/
asm	je	RX_INT
asm	cmp 	al,2			/*TX interrupt ?*/
asm	je	TX_INT
asm	clc	/*dummy instruction to speed up jmp instruction*/
asm	jmp	END_INT			/*prepare to leave interrupt*/
asm	clc	/*dummy instruction to speed up jmp instruction*/



/* RX interrupt handling -----------------------------------------------*/

RX_INT:
asm	mov	dx,RBR2			/*load RBR port adress*/
asm	in	al,dx			/*read received byte from UART*/

asm	les	bx,dword ptr inputfifo write_ptr /*load write pointer*/
asm	mov	byte ptr es:[bx],al	/*store received byte in fifo*/

asm	inc	word ptr inputfifo write_ptr	/*increment write pointer*/
asm	inc	word ptr inputfifo content	/*increment fifo content*/

asm	cmp	bx,word ptr inputfifo end /*wrap around ?*/
asm	je	rxm2			/*if yes handle it at rxm2*/

rxm1:
asm	mov	ax,inputfifo limit	/*load fill limit*/
asm	cmp	inputfifo content,ax /*compare fifo content vs limit*/
asm	jb	NEXT_IIR		/*if still enougth space go on*/

/* fifo full handling*/
asm	mov	dx,MCR2			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,DTR_off		/*mask HW handshake bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp short NEXT_IIR		/*see if there is another IRQ*/

/* wrap around handling*/
rxm2:
asm	mov	ax,word ptr inputfifo start /*fifo start addr. offs.->ax*/
asm	mov	word ptr inputfifo write_ptr,ax /*wrap around*/
asm	jmp short rxm1				/*jump back to continue*/



/* TX interrupt handling -----------------------------------------------*/

TX_INT:                             /*check outputfifo content*/
#ifdef _MSC_VER                     
asm   mov ax, 0                     /* ensure a word alignment for MSC  */
asm	cmp	outputfifo content, ax
#else
asm	cmp	outputfifo content, 0	
#endif

asm	je	txm2			/*if there is more to transmit*/
					/*transmit it, else set RTS off*/
asm	mov	dx,MSR2			/*load MSR port address*/
asm	in	al,dx			/*read MSR*/

asm	and	al,CTS_DSR		/*check HW handshake*/
asm	jne	txm3

asm	les	bx,dword ptr outputfifo read_ptr /*load address of next*/
							/*byte*/

asm	mov	al,byte ptr es:[bx]	/*read next byte from fifo*/

asm	mov	dx,THR2			/*load THR port address*/
asm	out	dx,al			/*write next byte to THR*/

asm	inc	word ptr outputfifo read_ptr	/*increment read pointer*/
asm	dec	word ptr outputfifo content	/*decrement fifo content*/

asm	cmp	bx,word ptr outputfifo end /*wrap around ?*/
asm	je	txm4			/*if yes handle it at txm4*/

txm1:
asm	mov	hw_hs_wait+2,0		/*set HW handshake wait flag*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/


/* end of transmission handling*/
txm2:
asm	mov	dx,MCR2			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,RTS_off		/*mask RTS bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/

/* missing HW handshake handling*/
txm3:
asm	mov	hw_hs_wait+2,1		/*set HW handshake wait flag*/
asm	jmp 	NEXT_IIR

/* wrap around handling*/
txm4:
asm	mov	ax,word ptr outputfifo start /*fifo start addr.offs.->ax*/
asm	mov	word ptr outputfifo read_ptr,ax /*wrap around*/
asm	jmp short txm1				/*jump back to continue*/


/* exit interrupt handling ---------------------------------------------*/

END_INT:
asm	mov	al,EOSI			/*load end of interrupt command*/
asm	out	MEOI,al			/*send end of int.-> master cont.*/
asm	mov	al,EOI2			/*load end of interrupt command*/
asm	out	SEOI,al			/*send end of int.-> slave cont.*/

asm	pop 	es			/*restore all used registers,*/
asm	pop 	ds			/*remember Flags are restored*/
asm	pop 	dx			/*automatically by the CPU*/
asm	pop 	bx
asm	pop 	ax

#ifdef _MSC_VER
   asm   pop   bp    /* restore the stack frame (changed BS)   */
#endif

asm	iret
}
#undef start
#undef end
#undef write_ptr
#undef read_ptr
#undef content
#undef size
#undef limit

#endif

/***************************************************************************
*
* ROUTINE: mecif3_isr
*
* WRITTEN BY: Juergen Barthel                         ON:24.02.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine is the MECIF interrupt service routine for port 4
*
****************************************************************************
* RETURNED VALUE: none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* inputfifo[]      r   pointer to inputfifo structure
* outputfifo[]     r   pointer to outputfifo structure
*
****************************************************************************/
#ifdef SLOW
static void interrupt	mecif3_isr(void)
{
C_SLAVE_PORT_ISR(3)
}
#else
#define start +66
#define end +70
#define write_ptr +74
#define read_ptr +78
#define content +82
#define size +84
#define limit +86

static void mecif3_isr(void)
{
asm	push	ax			/*save all used registers,*/
asm	push 	bx			/*remember Flags are saved*/
asm	push 	dx			/*automatically by the CPU*/
asm	push 	ds
asm	push 	es



/* load correct datasegment -> DS --------------------------------------*/

asm	mov 	dx,seg inputfifo
asm	mov 	ds,dx



/* identificate interrupt ----------------------------------------------*/

NEXT_IIR:
asm	mov 	dx,IIR3			/*load IIR port address*/
asm	in  	al,dx			/*read IIR*/
asm	and 	al,7
asm	cmp	al,4			/*RX interrupt ?*/
asm	je	RX_INT
asm	cmp 	al,2			/*TX interrupt ?*/
asm	je	TX_INT
asm	clc	/*dummy instruction to speed up jmp instruction*/
asm	jmp	END_INT			/*prepare to leave interrupt*/
asm	clc	/*dummy instruction to speed up jmp instruction*/



/* RX interrupt handling -----------------------------------------------*/

RX_INT:
asm	mov	dx,RBR3			/*load RBR port adress*/
asm	in	al,dx			/*read received byte from UART*/

asm	les	bx,dword ptr inputfifo write_ptr /*load write pointer*/
asm	mov	byte ptr es:[bx],al	/*store received byte in fifo*/

asm	inc	word ptr inputfifo write_ptr	/*increment write pointer*/
asm	inc	word ptr inputfifo content	/*increment fifo content*/

asm	cmp	bx,word ptr inputfifo end /*wrap around ?*/
asm	je	rxm2			/*if yes handle it at rxm2*/

rxm1:
asm	mov	ax,inputfifo limit	/*load fill limit*/
asm	cmp	inputfifo content,ax /*compare fifo content vs limit*/
asm	jb	NEXT_IIR		/*if still enougth space go on*/

/* fifo full handling*/
asm	mov	dx,MCR3			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,DTR_off		/*mask HW handshake bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp short NEXT_IIR		/*see if there is another IRQ*/

/* wrap around handling*/
rxm2:
asm	mov	ax,word ptr inputfifo start /*fifo start addr. offs.->ax*/
asm	mov	word ptr inputfifo write_ptr,ax /*wrap around*/
asm	jmp short rxm1				/*jump back to continue*/



/* TX interrupt handling -----------------------------------------------*/

TX_INT:                             /*check outputfifo content*/
#ifdef _MSC_VER                     
asm   mov ax, 0                     /* ensure a word alignment for MSC  */
asm	cmp	outputfifo content, ax
#else
asm	cmp	outputfifo content, 0	
#endif

asm	je	txm2			/*if there is more to transmit*/
					/*transmit it, else set RTS off*/
asm	mov	dx,MSR3			/*load MSR port address*/
asm	in	al,dx			/*read MSR*/

asm	and	al,CTS_DSR		/*check HW handshake*/
asm	jne	txm3

asm	les	bx,dword ptr outputfifo read_ptr /*load address of next*/
							/*byte*/

asm	mov	al,byte ptr es:[bx]	/*read next byte from fifo*/

asm	mov	dx,THR3			/*load THR port address*/
asm	out	dx,al			/*write next byte to THR*/

asm	inc	word ptr outputfifo read_ptr	/*increment read pointer*/
asm	dec	word ptr outputfifo content	/*decrement fifo content*/

asm	cmp	bx,word ptr outputfifo end /*wrap around ?*/
asm	je	txm4			/*if yes handle it at txm4*/

txm1:
asm	mov	hw_hs_wait+3,0		/*set HW handshake wait flag*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/


/* end of transmission handling*/
txm2:
asm	mov	dx,MCR3			/*load MCR port address*/
asm	in	al,dx			/*read MCR*/
asm	and	al,RTS_off		/*mask RTS bit*/
asm	out	dx,al			/*write new MCR to UART*/
asm	jmp	NEXT_IIR		/*see if there is another IRQ*/

/* missing HW handshake handling*/
txm3:
asm	mov	hw_hs_wait+3,1		/*set HW handshake wait flag*/
asm	jmp 	NEXT_IIR

/* wrap around handling*/
txm4:
asm	mov	ax,word ptr outputfifo start /*fifo start addr.offs.->ax*/
asm	mov	word ptr outputfifo read_ptr,ax /*wrap around*/
asm	jmp short txm1				/*jump back to continue*/


/* exit interrupt handling ---------------------------------------------*/


END_INT:
asm	mov	al,EOSI			/*load end of interrupt command*/
asm	out	MEOI,al			/*send end of int.-> master cont.*/
asm	mov	al,EOI3			/*load end of interrupt command*/
asm	out	SEOI,al			/*send end of int.-> slave cont.*/

asm	pop 	es			/*restore all used registers,*/
asm	pop 	ds			/*remember Flags are restored*/
asm	pop 	dx			/*automatically by the CPU*/
asm	pop 	bx
asm	pop 	ax

#ifdef _MSC_VER
   asm   pop   bp    /* restore the stack frame (changed BS)   */
#endif

asm	iret
}
#undef start
#undef end
#undef write_ptr
#undef read_ptr
#undef content
#undef size
#undef limit

#endif

/***************************************************************************
*
* ROUTINE:read_msg
*
* WRITTEN BY: Juergen Barthel                         ON:8.3.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine reads a complete message from the (&inputfifo[port])
*
****************************************************************************
* RETURNED VALUE: pointer to the message that has been read,
*		  NULL if no complete message was read
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* rm		  r/w  shared memory for receiving of messages
* rsm[]           r/w  pointer to the last write positon in rm
* inputfifo[]      r   pointer to inputfifo
* outputfifo[]     r   pointer to outputfifo
* length[]        r/w  pointer to the last valid length found for this port
* sync_flag[]     r/w  pointer to a flag
* lost_char[]	  r/w  counter of lost data
* hw_hs_wait[]    r/w  hardware handshake flag
*
****************************************************************************/
u_16 *read_msg(u_16 port)
{
Fifo8		*infifo_ptr;
u_16	      	*rmptr;
u_8		*trsm;
u_16		*len;
u_8		*sync;
u_8		msg_rx_flag = FALSE;
u_8		c;
u_16    	i = 0;
u_8		j;
u_8		k;
PortTbl		*port_ptr;
Session		**session_ptr_p;


infifo_ptr = &inputfifo[port];
len = &(length[port]);
sync = &(sync_flag[port]);
trsm = rsm[port];


rmptr = (u_16 *)rm[port];

/* check content before reading */
while (infifo_ptr->content)
    {
    /* if current length is 0 hunt for a new SYNC */
    if (*len == 0)
	{
	trsm = &(rm[port][0]);
	if (*trsm != SYNC_HD)
	    {
	    if (!(infifo_ptr->content))
		break;/* the while loop */
	    FIFO_READ(infifo_ptr,*trsm);
	    }

	/* if SYNC_HD was read, read the next byte else go on */
	if (*trsm == SYNC_HD)
	    {
	    if (!(infifo_ptr->content))
		break;/* the while loop */
	    FIFO_READ(infifo_ptr,*trsm);

	    /* if the byte after a SYNC is not FFh and *
	     * not SYNC_HD a new begin was found           */
	    if ((*trsm != 0XFF) && (*trsm != SYNC_HD))
		{
		*len = 1;
		trsm++;
		}
	    else
		{
		lost_char[port]+=2;
		}
	    }
	else
	    {
	    lost_char[port]++;
	    }
	}

    /* if a new beginning was found read the second byte */
    if (*len == 1)
	{

	if (!(infifo_ptr->content))
	    break;/* the while loop */
	FIFO_READ(infifo_ptr,*trsm);

	/* if the second byte is not a SYNC_HD and the message length  *
	 * that was read is valid set *len to messagelength else store *
	 * the byte in rm[0] and set length to 0 so it can be          *
	 * processed in the if condition above during the next turn    */

	if ((*trsm  != SYNC_HD)     &&
	    (*rmptr >= MIN_MSG_LENGTH) &&
	    (*rmptr <= MAX_MSG_LENGTH) &&
	    (!(*rmptr & 0X01)))
	    {
	    *len = *rmptr;
	    trsm++;
	    }
	else
	    {
	    rm[port][0] = *trsm;
	    *len = 0;
	    lost_char[port]+=2;
	    }
	}

    /* if a valid length exists start reading else go on */
    if (*len >= MIN_MSG_LENGTH)
	{
	for(i = (u_16) (trsm - (u_8 *)&(rm[port][0]));(i < *len) || (*sync == TRUE);i++)
	    {
	    /* if there is still a byte in the inputbuffer to read,read it  *
	     * else store the length and the position of the receive shared *
	     * memory pointer and return NULL                               */
	    if (!(infifo_ptr->content))
		{
		rsm[port] = trsm;
		return(NULL);
		}

	    FIFO_READ(infifo_ptr,*trsm);
	    if (*sync == FALSE)
		{
		/* SYNC_HD handling */
		if (*trsm == SYNC_HD)
		    *sync = TRUE;
		trsm++;
		if ((i == *len - 1) && (*sync == FALSE))
		    msg_rx_flag = TRUE;
		}
	    else
		{
		*sync = FALSE;
		i--;
		/* if the byte is not a FF a new beginning could have been *
		 * found so write this byte to positon 0 of rm, break the  *
		 * reading loop and if it is a SYNC_HD set length to 0 so  *
		 * the byte can be processed by the first if condition of  *
		 * the while loop or if it is not a SYNC_HD set length to  *
		 * 1 so it can be processed by the second if condition of  *
		 * the while loop else go on without incrementing trsm	   */
		if (*trsm != 0XFF)
		    {
		    rm[port][0] = *trsm;
		    if (*trsm == SYNC_HD)
			{
			*len = 0;
			lost_char[port] += i;
			}
		    else
			{
			*len = 1;
			trsm = &(rm[port][1]);
			lost_char[port] += (i + 1);
			}
		    break; /* the for loop */
		    }
		if (i == (*len - 1))
		    msg_rx_flag = TRUE;
		}
	    }
	if (msg_rx_flag == TRUE)
	    break;
	}
    }

/* if the transmission was stopped try to start again */
if (hw_hs_wait[port] == 1)
    {
    /* if CTS and DSR  then start transmission again */
    if (((inportb(MSR) & CTS_DSR) == CTS_DSR) && (outputfifo[port].content))
	{
	FIFO_READ((&outputfifo[port]),c);
	hw_hs_wait[port] = 0;
	outportb(THR,c);
	}
    }

/* check if handshake is off */
if ((inportb(MCR) & DTR_on) == 0)
    {
    /* check if handshake can be set to on */
    if ((infifo_ptr->size - infifo_ptr->content) > 3)
	{
	/* set handshake */
	disable();
	outportb(MCR,inportb(MCR) | DTR_on);
	enable();
	}
    }


/* Any lost characters before ? */
if (lost_char[port] != 0)
    {
    lost_char[port] = 0;

    /* tell that all active sessions on that port */
    port_ptr = confport_tbl;
    for (j=0; j<MAX_HW_PORTS; j++)
	{
	if ((port_ptr->valid == TRUE) && (port_ptr->port_num == port))
	    {
	    /* search port_id in session_tbl */
	    session_ptr_p = session_ptr;
	    for (k=0; k<MAX_SESSIONS; k++)
		{
		if (*session_ptr_p != NULL)
		    {
		    if ((*session_ptr_p)->port_id == port_ptr->port_id)
			rx_error(MESSAGE_LOST, (*session_ptr_p)->con_id);
		    }
		session_ptr_p++;
		}
	    }
	port_ptr++;
	}
    }


if (msg_rx_flag == TRUE)
    {
    /* return complete message! */
    *len = 0;
    rsm[port] = &(rm[port][0]);
    return(rmptr);
    }

rsm[port] = trsm;
return(NULL);
}

/****************************************************************************
*
* ROUTINE:write_msg
*
* WRITTEN BY: Juergen Barthel                         ON:8.3.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION:this routine writes a complete message to the outputbuffer
*
****************************************************************************
* RETURNED VALUE: SUCCESS if successfull,
*
*		  FAIL if the specified message has an incorrect length,
*
*                 OUT_OF_TXBUFFER if outputbuffer is full and therefor the
*		                  specified message was not written to it,
*
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* outputfifo[]     r   pointer to outputfifo
* hw_hs_wait[]    r/w  hardware handshake flag
*
****************************************************************************/
i_16 write_msg(u_16 port,u_16 *mptr)
{
Fifo8		*outfifo_ptr;
u_8  		*tsm;
u_8		c;
i_16  		i;
i_16  		len;
i_16  		nff = 0;


/* if there is enougth space in the outputbuffer and the length in the 	*
 * message is valid write SYNC_HD first then write the content of      	*
 * 'mptr'to the outputbuffer 						*/
if ( (*mptr < MIN_MSG_LENGTH) || (*mptr > MAX_MSG_LENGTH) || (*mptr & 0X01))
    return(FAIL);

len = *mptr;
outfifo_ptr = (&outputfifo[port]);
tsm = (u_8 *)mptr;

/* count # of SYNC_HD in content of mptr *
 * to find out the # of 0XFF to be added */
for (i = 0;i < len;i++)
    {
    if (*tsm++ == SYNC_HD)
	nff++;
    }

tsm = (u_8 *)mptr;
/* check if there is enougth space in the outputbuffer */
if (outfifo_ptr->content > (OUTBUFSIZE - (len + 1 + nff)))
    return(OUT_OF_TXBUFFER);

/* write complete message to outputbuffer */
FIFO_WRITE(outfifo_ptr,SYNC_HD);

for (i = 0;i < len;++i,++tsm)
    {
    FIFO_WRITE(outfifo_ptr,*tsm);

    /* if SYNC_HD transmitted add FFh */
    if (*tsm == SYNC_HD)
	{
	FIFO_WRITE(outfifo_ptr,0XFF);
	}
    }

/* if CTS and DSR are set and no transmission is *
 * in progress (RTS off) start transmission      */
if ((inportb(MSR) & CTS_DSR) == CTS_DSR)
    {
    if ((inportb(MCR) & RTS_on) == 0)
	{
	outportb(MCR,inportb(MCR) | RTS_on);
	FIFO_READ(outfifo_ptr,c);
	outportb(THR,c);
	}
    else if (hw_hs_wait[port] == 1)
	{
	FIFO_READ(outfifo_ptr,c);
	hw_hs_wait[port] = 0;
	outportb(THR,c);
	}
    return(SUCCESS);
    }
else
    {
    /* set hw_hs_wait return errorcode */
    hw_hs_wait[port] = 1;
    return(SUCCESS);
    }
}

/***************************************************************************
*
* ROUTINE: mecif_end
*
* WRITTEN BY: Juergen Barthel                         ON:27.2.91
* CHANGED BY:                                         ON:
*
****************************************************************************
*
* DESCRIPTION: sets the RS232 as it was before the programm started
*
****************************************************************************
* RETURNED VALUE: none
*
* EXTERNAL REFERENCES:
*
* Identifier      r/w  Comment
* --------------------------------------------------------------------------
* inputfifo[]      r   pointer to inputfifo structure
* outputfifo[]     r   pointer to outputfifo structure
* hw_hs_wait[]    r/w  hardware handshake flag
* int_vect[]       r   SIO interrupt vector numbers
* imr_addr[]       r   interrupt mask register address
* mask_int[]       r   mask to disable SIO interrupts
* unmask_int[]     r   mask to enable SIO interrupts
* old_IER[]        r   contains the old content of the IER
* old_MCR[]        r   contains the old content of the MCR
* old_LCR[]        r   contains the old content of the LCR
* old_msb[]        r   contains the old content of the DLM
* old_lsb[]        r   contains the old content of the DLL
****************************************************************************/
void mecif_end(u_16 port)
{
u_8		c;
u_8		i;
u_8     	lsr;

/* check if transmission was stopped, if so try to start it again */
while (inportb(MSR) & RTS_on)
    {
    if (hw_hs_wait[port] == 1)
	{
	/* if CTS and DSR then start transmission again */
	if (((inportb(MCR) & CTS_DSR) == CTS_DSR) &&
	    (outputfifo[port].content != 0))
	    {
	    FIFO_READ((&outputfifo[port]),c);
	    outportb(THR,c);
	    hw_hs_wait[port] = 0;
	    }
	}
    }

/* wait until transmitter is empty */
do
    {
    lsr = (u_8) (inportb(LSR) & 0X40);
    }
while (!lsr && ((inportb(MCR) & CTS_DSR) == CTS_DSR) );


outportb(imr_addr[port],inportb(imr_addr[port]) | mask_int[port]);
outportb(LCR,0X80);
outportb(DLL,old_lsb[port]);
outportb(DLM,old_msb[port]);
outportb(LCR,old_LCR[port]);
outportb(IER,0X00);
outportb(MCR,0X00);
restore_old_isr(port);
outportb(IER,old_IER[port]);
outportb(MCR,old_MCR[port]);
for (i = 0;i <= 6;i++)               /* dummy read of SIO registers */
    inportb(sio[port] + i);
outportb(imr_addr[port],inportb(imr_addr[port]) & unmask_int[port]);


free(inputfifo[port].start);
free(outputfifo[port].start);

}
